'use strict';

const faker = require('@faker-js/faker');
const typeAnalyzer = require('@previewjs/type-analyzer');
const assertNever = require('assert-never');
const ts = require('typescript');
const prettier = require('prettier');
const parserBabel = require('prettier/parser-babel.js');

const RESERVED_IDENTIFIERS = /* @__PURE__ */ new Set(["arguments", "get", "set"]);
function isValidPropName(name) {
  return /^[a-zA-Z_$][a-zA-Z_$0-9]*$/.test(name) && !RESERVED_IDENTIFIERS.has(name);
}

function array(items) {
  return {
    kind: "array",
    items
  };
}
const EMPTY_ARRAY = array([]);
function boolean(value) {
  return {
    kind: "boolean",
    value
  };
}
const FALSE = boolean(false);
const TRUE = boolean(true);
function fn(returnValue) {
  return {
    kind: "function",
    returnValue
  };
}
function map(values) {
  return {
    kind: "map",
    values
  };
}
const NULL = { kind: "null" };
function number(value) {
  return {
    kind: "number",
    value
  };
}
function object(entries) {
  return {
    kind: "object",
    entries
  };
}
function promise(value) {
  return {
    kind: "promise",
    value
  };
}
const EMPTY_OBJECT = object([]);
const EMPTY_MAP = map(EMPTY_OBJECT);
function set(values) {
  return {
    kind: "set",
    values
  };
}
const EMPTY_SET = set(EMPTY_ARRAY);
function string(value) {
  return {
    kind: "string",
    value
  };
}
const UNDEFINED = { kind: "undefined" };
function unknown(source) {
  return {
    kind: "unknown",
    source
  };
}
const UNKNOWN = unknown();

function generateSerializableValue(type, collected, options = {}) {
  return _generateSerializableValue(
    type,
    collected,
    options.fieldName || "",
    [],
    options.random || false,
    false
  );
}
function _generateSerializableValue(type, collected, fieldName, rejectTypeNames, random, isFunctionReturnValue) {
  let encounteredAliases;
  [type, encounteredAliases] = typeAnalyzer.dereferenceType(
    type,
    collected,
    rejectTypeNames
  );
  rejectTypeNames = [...rejectTypeNames, ...encounteredAliases];
  switch (type.kind) {
    case "any":
    case "unknown":
    case "never":
    case "void":
      return UNDEFINED;
    case "null":
      return NULL;
    case "boolean":
      return random && Math.random() < 0.5 ? TRUE : FALSE;
    case "string":
    case "node":
      return string(
        random ? faker.faker.lorem.words(generateRandomInteger(0, 10)) : fieldName
      );
    case "number":
      return number(random ? generateRandomInteger() : 0);
    case "literal":
      if (typeof type.value === "number") {
        return number(type.value);
      } else if (typeof type.value === "string") {
        return string(type.value);
      } else {
        return type.value ? TRUE : FALSE;
      }
    case "enum": {
      const optionValues = Object.values(type.options);
      const value = optionValues[random ? generateRandomInteger(0, optionValues.length) : 0];
      if (typeof value === "number") {
        return number(value);
      } else {
        return string(value || "unknown");
      }
    }
    case "array":
      return generateArrayValue(
        type,
        collected,
        fieldName,
        rejectTypeNames,
        random,
        isFunctionReturnValue
      );
    case "set":
      return set(
        generateArrayValue(
          typeAnalyzer.arrayType(type.items),
          collected,
          fieldName,
          rejectTypeNames,
          random,
          isFunctionReturnValue
        )
      );
    case "tuple":
      return array(
        type.items.map(
          (item) => _generateSerializableValue(
            item,
            collected,
            fieldName,
            rejectTypeNames,
            random,
            isFunctionReturnValue
          )
        )
      );
    case "object": {
      const entries = [];
      for (const [propName, propType] of Object.entries(type.fields)) {
        const propValue = _generateSerializableValue(
          propType,
          collected,
          propName,
          rejectTypeNames,
          random,
          isFunctionReturnValue
        );
        if (propValue.kind === "undefined") {
          continue;
        }
        if (!isValidPropName(propName)) {
          continue;
        }
        entries.push({
          key: string(propName),
          value: propValue
        });
      }
      return object(entries);
    }
    case "map":
      return map(
        generateRecordValue(
          {
            ...type,
            kind: "record"
          },
          collected,
          fieldName,
          rejectTypeNames,
          random,
          isFunctionReturnValue
        )
      );
    case "record":
      return generateRecordValue(
        {
          ...type,
          kind: "record"
        },
        collected,
        fieldName,
        rejectTypeNames,
        random,
        isFunctionReturnValue
      );
    case "union":
      if (!random) {
        if (typeAnalyzer.isValid(type, collected, void 0)) {
          return UNDEFINED;
        }
        if (typeAnalyzer.isValid(type, collected, null)) {
          return NULL;
        }
        if (typeAnalyzer.isValid(type, collected, false)) {
          return FALSE;
        }
      }
      return _generateSerializableValue(
        type.types[random ? generateRandomInteger(0, type.types.length) : 0],
        collected,
        fieldName,
        rejectTypeNames,
        random,
        isFunctionReturnValue
      );
    case "intersection":
      return _generateSerializableValue(
        type.types[0],
        collected,
        fieldName,
        rejectTypeNames,
        random,
        isFunctionReturnValue
      );
    case "function":
      return fn(
        isFunctionReturnValue ? UNDEFINED : _generateSerializableValue(
          type.returnType,
          collected,
          fieldName,
          rejectTypeNames,
          random,
          true
        )
      );
    case "optional":
      if (random && Math.random() < 0.5) {
        return _generateSerializableValue(
          type.type,
          collected,
          fieldName,
          rejectTypeNames,
          random,
          isFunctionReturnValue
        );
      }
      return UNDEFINED;
    case "promise": {
      return promise({
        type: "reject",
        message: null
      });
    }
    case "name":
      return _generateSerializableValue(
        type,
        collected,
        fieldName,
        rejectTypeNames,
        random,
        isFunctionReturnValue
      );
    default:
      throw assertNever.assertNever(type);
  }
}
function generateArrayValue(type, collected, fieldName, rejectTypeNames, random, isFunctionReturnValue) {
  if (isFunctionReturnValue) {
    return EMPTY_ARRAY;
  }
  const itemValues = [];
  const length = random ? generateRandomInteger(0, 3) : 1;
  for (let i = 0; i < length; i++) {
    const itemValue = _generateSerializableValue(
      type.items,
      collected,
      fieldName,
      rejectTypeNames,
      random,
      isFunctionReturnValue
    );
    itemValues.push(itemValue);
  }
  if (itemValues.every(
    (itemValue) => itemValue.kind === "undefined" || itemValue.kind === "object" && Object.keys(itemValue.entries).length === 0
  )) {
    return EMPTY_ARRAY;
  }
  return array(itemValues);
}
function generateRecordValue(type, collected, fieldName, rejectTypeNames, random, isFunctionReturnValue) {
  if (!random) {
    return EMPTY_OBJECT;
  }
  const values = generateArrayValue(
    {
      kind: "array",
      items: type.values
    },
    collected,
    fieldName,
    rejectTypeNames,
    random,
    isFunctionReturnValue
  ).items;
  const entries = [];
  for (const value of values) {
    const key = _generateSerializableValue(
      type.keys,
      collected,
      fieldName,
      rejectTypeNames,
      random,
      isFunctionReturnValue
    );
    entries.push({
      key,
      value
    });
  }
  return object(entries);
}
function generateRandomInteger(minInclusive = -5e3, maxExclusive = 5e3) {
  const range = maxExclusive - minInclusive;
  return Math.floor(Math.random() * range + minInclusive);
}

function parseSerializableValue(expression) {
  const fallbackValue = unknown(expression.getText());
  if (ts.isParenthesizedExpression(expression)) {
    return parseSerializableValue(expression.expression);
  }
  if (expression.kind === ts.SyntaxKind.NullKeyword) {
    return NULL;
  }
  if (expression.kind === ts.SyntaxKind.TrueKeyword) {
    return TRUE;
  }
  if (expression.kind === ts.SyntaxKind.FalseKeyword) {
    return FALSE;
  }
  if (expression.kind === ts.SyntaxKind.UndefinedKeyword || ts.isIdentifier(expression) && expression.text === "undefined") {
    return UNDEFINED;
  }
  if (ts.isNumericLiteral(expression)) {
    try {
      return number(parseFloat(expression.text));
    } catch {
      return fallbackValue;
    }
  }
  if (ts.isPrefixUnaryExpression(expression) && expression.operator === ts.SyntaxKind.MinusToken && ts.isNumericLiteral(expression.operand)) {
    try {
      return number(-parseFloat(expression.operand.text));
    } catch {
      return fallbackValue;
    }
  }
  if (ts.isStringLiteral(expression)) {
    return string(expression.text);
  }
  if (ts.isNoSubstitutionTemplateLiteral(expression)) {
    return string(expression.text);
  }
  if (ts.isArrayLiteralExpression(expression)) {
    return array(expression.elements.map(parseSerializableValue));
  }
  if (ts.isNewExpression(expression) && (expression.arguments?.length || 0) <= 1) {
    if (ts.isIdentifier(expression.expression)) {
      if (expression.expression.text === "Set") {
        const firstArgument = expression.arguments && expression.arguments[0];
        if (!firstArgument) {
          return EMPTY_SET;
        } else if (ts.isArrayLiteralExpression(firstArgument)) {
          return set(array(firstArgument.elements.map(parseSerializableValue)));
        }
      }
      if (expression.expression.text === "Map") {
        const firstArgument = expression.arguments && expression.arguments[0];
        if (!firstArgument) {
          return EMPTY_MAP;
        } else if (ts.isArrayLiteralExpression(firstArgument)) {
          const entries = [];
          for (const element of firstArgument.elements) {
            if (!ts.isArrayLiteralExpression(element) || element.elements.length !== 2) {
              return fallbackValue;
            }
            entries.push({
              key: parseSerializableValue(element.elements[0]),
              value: parseSerializableValue(element.elements[1])
            });
          }
          return map(object(entries));
        } else if (ts.isCallExpression(firstArgument) && ts.isPropertyAccessExpression(firstArgument.expression) && ts.isIdentifier(firstArgument.expression.expression) && firstArgument.expression.expression.text === "Object" && firstArgument.expression.name.text === "entries" && firstArgument.arguments.length === 1) {
          const object2 = parseSerializableValue(firstArgument.arguments[0]);
          if (object2.kind !== "object") {
            return fallbackValue;
          }
          return map(object2);
        }
      }
    }
  }
  if (ts.isCallExpression(expression) && ts.isPropertyAccessExpression(expression.expression) && ts.isIdentifier(expression.expression.expression) && expression.expression.expression.text === "Promise" && ts.isIdentifier(expression.expression.name) && expression.arguments.length <= 1) {
    const firstArgument = expression.arguments[0];
    if (expression.expression.name.text === "resolve") {
      const value = firstArgument ? parseSerializableValue(firstArgument) : UNDEFINED;
      return promise({
        type: "resolve",
        value
      });
    }
    if (expression.expression.name.text === "reject") {
      if (!firstArgument) {
        return promise({
          type: "reject",
          message: null
        });
      } else if (ts.isNewExpression(firstArgument) && (firstArgument.arguments?.length || 0) <= 1) {
        const firstErrorArgument = firstArgument.arguments && firstArgument.arguments[0];
        if (firstErrorArgument) {
          if (ts.isStringLiteral(firstErrorArgument) || ts.isNoSubstitutionTemplateLiteral(firstErrorArgument)) {
            return promise({
              type: "reject",
              message: firstErrorArgument.text
            });
          }
        }
      }
    }
  }
  if (ts.isObjectLiteralExpression(expression)) {
    const entries = [];
    for (const property of expression.properties) {
      if (ts.isShorthandPropertyAssignment(property)) {
        entries.push({
          key: string(property.name.text),
          value: UNKNOWN
        });
      } else if (ts.isPropertyAssignment(property) && (ts.isIdentifier(property.name) || ts.isStringLiteral(property.name) || ts.isNumericLiteral(property.name))) {
        entries.push({
          key: string(property.name.text),
          value: parseSerializableValue(property.initializer)
        });
      } else {
        return fallbackValue;
      }
    }
    return object(entries);
  }
  if (ts.isArrowFunction(expression) && expression.parameters.length === 0) {
    if (ts.isBlock(expression.body)) {
      const statements = expression.body.statements;
      if (statements.length === 0) {
        return fn(UNDEFINED);
      } else if (statements.length === 1) {
        const returnStatement = statements[0];
        if (!ts.isReturnStatement(returnStatement)) {
          return fallbackValue;
        }
        return fn(
          returnStatement.expression ? parseSerializableValue(returnStatement.expression) : UNDEFINED
        );
      } else {
        return fallbackValue;
      }
    } else {
      return fn(parseSerializableValue(expression.body));
    }
  }
  if (ts.isFunctionExpression(expression) && expression.parameters.length === 0) {
    const statements = expression.body.statements;
    if (statements.length === 0) {
      return fn(UNDEFINED);
    } else if (statements.length === 1) {
      const returnStatement = statements[0];
      if (!ts.isReturnStatement(returnStatement)) {
        return fallbackValue;
      }
      return fn(
        returnStatement.expression ? parseSerializableValue(returnStatement.expression) : UNDEFINED
      );
    } else {
      return fallbackValue;
    }
  }
  return fallbackValue;
}

function serializableValueToJavaScript(value) {
  let expression = serializableValueToUnformattedJavaScript(value);
  try {
    expression = formatExpression(expression);
  } catch {
  }
  return expression;
}
function formatExpression(expressionSource) {
  const formattedStatement = prettier.format(`value = ${expressionSource}`, {
    parser: "babel",
    plugins: [parserBabel],
    filepath: "component.js",
    trailingComma: "none"
  }).trim();
  return formattedStatement.replace(/^value = ((.|\s)*);$/m, "$1");
}
function serializableValueToUnformattedJavaScript(value) {
  switch (value.kind) {
    case "array":
      return `[${value.items.map((item) => serializableValueToUnformattedJavaScript(item)).join(", ")}]`;
    case "boolean":
      return value.value ? "true" : "false";
    case "function":
      return `() => ${value.returnValue.kind === "undefined" ? "{}" : `(${serializableValueToUnformattedJavaScript(value.returnValue)})`}`;
    case "map":
      return `new Map(${value.values.entries.length > 0 ? `Object.entries(${serializableValueToUnformattedJavaScript(
        value.values
      )})` : ""})`;
    case "null":
      return "null";
    case "number":
      return value.value.toString(10);
    case "object": {
      if (Object.entries(value.entries).length === 0) {
        return "{}";
      }
      let text = "";
      text += "{\n";
      for (const entry of value.entries) {
        text += `${entry.key.kind === "string" ? JSON.stringify(entry.key.value) : `[${serializableValueToUnformattedJavaScript(entry.key)}]`}: ${serializableValueToUnformattedJavaScript(entry.value)},
`;
      }
      text += "\n}";
      return text;
    }
    case "promise":
      if (value.value.type === "reject") {
        return `Promise.reject(${value.value.message === null ? "" : `new Error(${JSON.stringify(value.value.message)})`})`;
      } else {
        return `Promise.resolve(${serializableValueToUnformattedJavaScript(
          value.value.value
        )})`;
      }
    case "set":
      return `new Set(${value.values.items.length > 0 ? serializableValueToUnformattedJavaScript(value.values) : ""})`;
    case "string":
      return JSON.stringify(value.value);
    case "undefined":
      return "undefined";
    case "unknown": {
      const source = value.source ?? "{}";
      try {
        formatExpression(source);
        return source;
      } catch {
        return "{}";
      }
    }
    default:
      throw assertNever(value);
  }
}

exports.EMPTY_ARRAY = EMPTY_ARRAY;
exports.EMPTY_MAP = EMPTY_MAP;
exports.EMPTY_OBJECT = EMPTY_OBJECT;
exports.EMPTY_SET = EMPTY_SET;
exports.FALSE = FALSE;
exports.NULL = NULL;
exports.TRUE = TRUE;
exports.UNDEFINED = UNDEFINED;
exports.UNKNOWN = UNKNOWN;
exports.array = array;
exports.boolean = boolean;
exports.fn = fn;
exports.generateSerializableValue = generateSerializableValue;
exports.map = map;
exports.number = number;
exports.object = object;
exports.parseSerializableValue = parseSerializableValue;
exports.promise = promise;
exports.serializableValueToJavaScript = serializableValueToJavaScript;
exports.set = set;
exports.string = string;
exports.unknown = unknown;
